#include "opendefs.h"
#include "urandom.h"
#include "openqueue.h"
#include "openserial.h"
#include "packetfunctions.h"
#include "scheduler.h"
#include "IEEE802154E.h"
#include "idmanager.h"
#include "time.h"
#include "stdlib.h"

//=========================== variables =======================================

urandom_vars_send_t urandom_vars_send;
urandom_vars_receive_t urandom_vars_receive;

static const uint8_t urandom_dst_addr[]   = {
   0xbb, 0xbb, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01
};

//=========================== prototypes ======================================

void urandom_timer_cb(opentimers_id_t id);
void urandom_task_cb(void);
void urandom_timer_init(urandom_vars_t *send, uint16_t period_aux);
void urandom_register_socket(urandom_vars_t *urandom, bool send_receive);

//=========================== public ==========================================

void urandom_init(void) {
   urandom_register_socket(&urandom_vars_send, TRUE);
   urandom_register_socket(&urandom_vars_receive, FALSE);
   urandom_timer_init(&urandom_vars_send, URANDOM_PERIOD_MS);
}

void urandom_sendDone(OpenQueueEntry_t* msg, owerror_t error) {
   openqueue_freePacketBuffer(msg);
}

void urandom_receive(OpenQueueEntry_t* pkt) {

   opentimers_destroy(urandom_vars_send.timerId);

   if((URANDOM_PERIOD_MS >= atoi((char *)&pkt->payload[0]))){
       urandom_timer_init(&urandom_vars_send, atoi((char *)&pkt->payload[0]));
   } else {
       urandom_timer_init(&urandom_vars_send, URANDOM_PERIOD_MS);
       return;
   }

   uint16_t          temp_l4_destination_port;
   OpenQueueEntry_t* reply;

   reply = openqueue_getFreePacketBuffer(COMPONENT_URANDOM);
   if (reply==NULL) {
      openserial_printError(
         COMPONENT_URANDOM,
         ERR_NO_FREE_PACKET_BUFFER,
         (errorparameter_t)0,
         (errorparameter_t)0
      );
      openqueue_freePacketBuffer(pkt); //clear the request packet as well
      return;
   }

   reply->owner                         = COMPONENT_URANDOM;

   // reply with the same OpenQueueEntry_t
   reply->creator                       = COMPONENT_URANDOM;
   reply->l4_protocol                   = IANA_UDP;
   temp_l4_destination_port             = pkt->l4_destination_port;
   reply->l4_destination_port           = pkt->l4_sourcePortORicmpv6Type;
   reply->l4_sourcePortORicmpv6Type     = temp_l4_destination_port;
   reply->l3_destinationAdd.type        = ADDR_128B;

   // copy source to destination to echo.
   memcpy(&reply->l3_destinationAdd.addr_128b[0],&pkt->l3_sourceAdd.addr_128b[0],16);
   uint8_t message_payload[]    = "OK, period is changed";
   packetfunctions_reserveHeaderSize(reply,sizeof(message_payload)-1);
   memcpy(&reply->payload[0],message_payload,sizeof(message_payload)-1);

   openqueue_freePacketBuffer(pkt);

   if ((openudp_send(reply))==E_FAIL) {
      openqueue_freePacketBuffer(reply);
   }
}

//=========================== private =========================================

/**
\note timer fired, but we don't want to execute task in ISR mode instead, push
   task to scheduler with CoAP priority, and let scheduler take care of it.
*/
void urandom_timer_cb(opentimers_id_t id){

   scheduler_push_task(urandom_task_cb,TASKPRIO_COAP);
}

void urandom_task_cb(void) {
   OpenQueueEntry_t*    pkt;

   // don't run if not synch
   if (ieee154e_isSynch() == FALSE) return;

   // don't run on dagroot
   if (idmanager_getIsDAGroot()) {
      opentimers_destroy(urandom_vars_send.timerId);
      return;
   }

   // if you get here, send a packet

   // get a free packet buffer
   pkt = openqueue_getFreePacketBuffer(COMPONENT_URANDOM);
   if (pkt==NULL) {
      openserial_printError(
          COMPONENT_URANDOM,
          ERR_NO_FREE_PACKET_BUFFER,
          (errorparameter_t)0,
          (errorparameter_t)0
      );
      return;
   }

   pkt->owner                         = COMPONENT_URANDOM;
   pkt->creator                       = COMPONENT_URANDOM;
   pkt->l4_protocol                   = IANA_UDP;
   pkt->l4_destination_port           = WKP_UDP_RANDOM_SEND;
   pkt->l4_sourcePortORicmpv6Type     = WKP_UDP_RANDOM_SEND;
   pkt->l3_destinationAdd.type        = ADDR_128B;
   memcpy(&pkt->l3_destinationAdd.addr_128b[0],urandom_dst_addr,16);

   // generate random number between 0 and 100
   uint8_t low_num = 0, hi_num = 101;
   srand(time(NULL));
   uint16_t random_number= (uint16_t)(rand() % (hi_num - low_num)) + low_num;
   
   // add payload
   packetfunctions_reserveHeaderSize(pkt,sizeof(uint16_t));
   pkt->payload[1] = (uint8_t)((random_number & 0xff00)>>8);
   pkt->payload[0] = (uint8_t)(random_number & 0x00ff);

   if ((openudp_send(pkt))==E_FAIL) {
      openqueue_freePacketBuffer(pkt);
   }
}

void urandom_register_socket(urandom_vars_t *urandom, bool send_receive){

   memset(urandom,0,sizeof(urandom_vars_t));

   // register at UDP stack
   if(send_receive){
      urandom->desc.port              = WKP_UDP_RANDOM_SEND;
      urandom->desc.callbackSendDone  = NULL;
      urandom->desc.callbackReceive   = NULL;
   } else {
      urandom->desc.port              = WKP_UDP_RANDOM_RECEIVE;
      urandom->desc.callbackSendDone  = &urandom_sendDone;
      urandom->desc.callbackReceive   = &urandom_receive;
   }
   openudp_register(&urandom->desc);
}

void urandom_timer_init(urandom_vars_t *send, uint16_t period_aux){
  send->period = period_aux;
  // start periodic timer
  send->timerId = opentimers_create();
  opentimers_scheduleIn(
      send->timerId,
      period_aux,
      TIME_MS,
      TIMER_PERIODIC,
      urandom_timer_cb
  );
}
